#!/usr/bin/env python
# Copyright (c) 2011
# All rights reserved.
# Original author: Michal T. Lorenc m.t.lorenc@wp.pl
# Small fixes in 1.6: Philipp Bayer philipp.bayer@uqconnect.edu.au
# What was fixed: 
# Error when only 1 reference was used, but user supplied multiple CPUs - integer/string concatenation led to crash
# Program now exits when no bam-index (*bam.bai) is found
# Still left: chr_output and contig_output generate identical files
# Addition in 1.8: Philipp Bayer
# Some code-cleanup, additional comments, optional reporting of heterozygous SNPs, removed BioPython-dependency for speed, all output now logged to std.err
# Addition in 1.9: Philipp Bayer
# Fixes bash-related bug with ' or "
# Addition in 2.0: PHilipp Bayer
# Added Kaitao's/Michal's vcf 4.1
# Addition in 2.001: Kenneth Chan
# Added default eone to ref_base in SNP_info constructor so it's backward compatible with create_flapjack_files_v2.py 

# Brassica usage:
# python SGSautoSNP.py
# --bam /media/BIGMICHAL/tests/\
#        J_E_A_S_M1_M2_Bn_Sr_cloneremoved_vs_XA10m_v3.0.bam
# --fasta /media/BIGMICHAL/tests/XA10m_v3.0_small.fasta
# --chr_offset /media/BIGMICHAL/tests/XA10_v3.0_contig.gff3
# --cultivars 'J,E,A,S,M1,M2,Bn,Sr'
# --snp_id_prefix UQ01BC
# --pl UQ_BNSNP
# --contig_output /media/BIGMICHAL/tests/\
#        J_E_A_S_M1_M2_Bn_Sr_cloneremoved_vs_XA10m_v3.0_contig_output
# --chr_output /media/BIGMICHAL/tests/\
#        J_E_A_S_M1_M2_Bn_Sr_cloneremoved_vs_XA10m_v3.0_chr_output
# --cpu 1

# Wheat usage:
# python SGSautoSNP.py
# --bam D_E_vs_7A_SynBuild_v2.0_combined_7A_SynBuild_v2.0.extr_cont.m.bam
# --fasta 7A_SynBuild_v2.0_combined_7A_SynBuild_v2.0.extra_contigs.m.fasta
# --cultivars 'D,E,G,R'
# --contig_output D_E_vs_7A_SynBuild_v2.0_combined_7A_SynBuild_v2.0.extr_cont.m
# --cpu 15


import re
import os
import sys
import pysam
import math
import datetime
from multiprocessing import Process, JoinableQueue, Queue
import argparse

# Global constants
MIN_SCORE    = 2    # A SNP needs to have at least MIN_SCORE alleles
MIN_COVERAGE = 4    # A SNP needs to have at least MIN_COVERAGE reads align at that position
VERSION      = 2.001# Version number

class Base_info():
    """ 
    Records basic information about a single base.

    Args:
        pos: Position in nucleotides.
        base: Base at the position (ATGC)
        cultivar: Cultivar for which this base was recorded. Example: T for Tapidor.

    """

    def __init__(self, pos, base, cultivar):
        self.pos = pos
        self.base = base
        self.cultivar = cultivar

    def __repr__(self):
        return self.cultivar + "=" + self.base + "=" + str(self.pos) + ";"


class Genotype_info():
    """ 
    Records information about bases of a genotype.

    Args:
        base: The nucleotide (A,T,G,C)
        amount: How often that nucleotide appears (most often 1)
    
    Changes:
        self.bases: A dictionary counting how often each base appears for a given genotype

    """

    def __init__(self, base, amount):
        self.bases = {base:amount}
    def add(self, base, amount):
        if base in self.bases:
            self.bases[base] += amount
        else:
            self.bases[base] = amount


def get_id_names(fasta):
    """
    Standin function for SeqIO.parse() - since we need only the names of the sequences, it's faster to iterate through
    the file instead of creating Seq-Objects

    Args:
        fasta: name of the file to be parsed
    Returns:
        seqs: a list of seq-names with the ">" removed

    """
    seqs = []
    for line in open(fasta):
        if line[0] == ">":
            # we want the ID, not the full description! Split by spaces and take the first element
            name = line.rstrip().split(" ")[0].replace(">","")
            seqs.append(name)
    return seqs

class SNP_info():
    """
    Stores all relevant information about a SNP.

    Args:
        contig_pos: position on contigs (if they exist)
        chr_pos: position on chromosome (identical to contig_pos in some cases)
        snp_score: score between 2 and infinity
        genotypes: dictionary of type Genotype_info, stores genotypes for all cultivars
        allele: all possible alleles on this SNP ("A/T", or "A/T/G")

    """

    def __init__(self, contig_pos, chr_pos, snp_score, genotypes, allele, ref_base=None):
        self.contig_pos = contig_pos
        self.chr_pos = chr_pos
        self.snp_score = snp_score
        self.genotypes = genotypes
        self.allele = allele
        self.snp_id = ''
        self.ref_base = ref_base
    
    def __repr__(self):
        return "Position on contig: %s\nPosition on chromosome: %s\nScore: %s\nGenotypes: %s\nAlleles: %s\n" \
                %(self.contig_pos, self.chr_pos, self.snp_score, self.genotypes, self.allele)


class SNPcaller(Process):
    
    def __init__(self, task_queue, result_queue, bam_filename, cultivars, offsets,
     			 heterozygous, ref_fasta):
        Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue
        self.bam_fh = pysam.Samfile(bam_filename, "rb")
        self.cultivars = cultivars
        self.offsets = offsets
        self.heterozygous = heterozygous
        self.ref_fasta = ref_fasta

    def run(self):
        #proc_name = self.name
        fasta_filename_fh = pysam.Fastafile(self.ref_fasta)
        while True:
            ref_name = self.task_queue.get()
            if ref_name is None:# None means shutdown
                self.task_queue.task_done()
                break
            
            contig_seq = fasta_filename_fh.fetch(ref_name)
            status = self.__sgsAutoSNP(ref_name, contig_seq)
            if status != 0:
                self.task_queue.task_done()
                break

        self.bam_fh.close()

    def __sgsAutoSNP(self, ref_name, contig_seq):
        # Record all letters before a "P" or a "S"
        # Example: TP1 becomes T
        #          OachkatzelschwoafS1 becomes Oachkatzelschwoaf
        cultivar_regex = re.compile('^([^-]+)(?:P|S)\d')
        try:
            chr_start_pos = self.offsets[ref_name]
        except KeyError:
            sys.stderr.write("Error: Couldn't find " + ref_name + " in offsets-file. Check your gff3.\n Offsets are: %s" %self.offsets.keys())
            self.result_queue.put((ref_name, False))
            return 1

        snps = []  
        # iterate over each position in the bam-file
        for pileupcolumn in self.bam_fh.pileup(ref_name):
            base_infos = []
            # do we have enough reads align at this position?
            if pileupcolumn.n >= MIN_COVERAGE:
                for pileupread in pileupcolumn.pileups:
                    cultivar = cultivar_regex.findall(pileupread.alignment.qname)[0]
                    pos = pileupcolumn.pos
                    base = pileupread.alignment.seq[pileupread.qpos].upper()
                    base_infos.append(Base_info(pos, base, cultivar))
                    
                snp_score = self.__calc_snp_score(base_infos)

                # if we want to report heterozygous SNPs, we skip checking for conflicting SNPs
                if self.heterozygous:
                    # check whether we only have one genotype at this position
                    is_conflict = self.__is_monomorphic(base_infos)
                else:
                    is_conflict = self.__is_conflict(base_infos)

                # Is this a valid SNP?
                if snp_score >= MIN_SCORE and not is_conflict:
                    contig_pos = pileupcolumn.pos
                    chr_pos = chr_start_pos + contig_pos
                    genotypes = self.__find_genotypes(cultivars, base_infos)
                    allele = self.__find_allele(genotypes)
                    #print ref_base
                    ref_base = contig_seq[contig_pos]
                    s = SNP_info(contig_pos, chr_pos, snp_score, genotypes, allele, ref_base)
                    snps.append(s)

        # report that we're done
        self.task_queue.task_done()
        # append our result to all results
        self.result_queue.put((ref_name, snps))
        return 0
  
    def __is_monomorphic(self, base_infos):
        cultivars = {}
    
        for read in base_infos:
            cultivar = read.cultivar
            base = read.base
            if cultivar not in cultivars:
                cultivars[cultivar] = base
            else:
                cultivars[cultivar] += base

        # do we have only one cultivar here?
        if len(cultivars.keys()) <= 1:
            return True

        # "union" all alleles we have
        set_of_alleles = set()
        for cultivar in cultivars:
            set_of_alleles = set_of_alleles.union(set(cultivars[cultivar]))
        # is this monomorphic?
        if len(set_of_alleles) <= 1:
            return True

        return False
 

    def __find_allele(self, genotypes):
        """
        Returns a string showing all present alleles.

        Returns:
            alleles: A string with all alleles in the form "A/T/G/C"

        """
        
        allele = {}
    
        for genotype in genotypes.values():
            bases = genotype.bases
            for base in bases.keys():
                if base != 'X':
                    allele[base] = 1
        alleles = '/'.join(sorted(allele.keys()))

        return alleles
    
    
    def __find_genotypes(self, cultivars, base_infos):
        genotypes = {}
    
        for read in base_infos:
            cultivar = read.cultivar
            base = read.base
    
            if cultivar not in genotypes:
                genotypes[cultivar] = Genotype_info(base, 1)
            else:
                genotypes[cultivar].add(base, 1)

        # make sure we saw all possible cultivars
        for cultivar in cultivars:
            # if we didn't saw a cultivar, give that one the allele X with count 0
            if cultivar not in genotypes:
                genotypes[cultivar] = Genotype_info('X', 0)
        return genotypes
    
    
    def __is_conflict(self, base_infos):
        cultivars = {}
    
        for read in base_infos:
            cultivar = read.cultivar
            base = read.base
            if cultivar not in cultivars:
                cultivars[cultivar] = base
            else:
                cultivars[cultivar] += base
    
        for cul, bases in cultivars.items():
            # heterozygous SNP?
            if len({}.fromkeys(bases).keys()) > 1:
                return True
        return False
    
    
    def __calc_snp_score(self, base_infos):
        """ 

        Reports the SNP score, which is the count of the allele that appears the least.

        Example: A 5, C 2, G 2, then the SNP score is 2 because C or G appear only 2 times.
                 A 10, then the SNP score is 10 because A appears 10 times.

        Returns:
            min_value (int): The SNP score.
        """
        score = {}
    
        for read in base_infos:
            base = read.base
            if base not in score:
                score[base] = 1
            else:
                score[base] += 1
        min_value = min(score.values())
    
        if min_value < len(base_infos):
            return min_value
        else:
            return 0


def read_offset_file(offset_filename):
    offsets = {}
    with open(offset_filename) as f:
        f.next()
        f.next()
        for line in f:
            if line.strip():
                splitted_line = line.rstrip().split('\t')
                contig_name = re.findall('\AID=([^;]+);', splitted_line[8])[0]
                chr_start_pos = int(splitted_line[3]) - 1
                offsets[contig_name] = chr_start_pos
        chr_name = splitted_line[0]
    return (offsets, chr_name)

def create_fake_offset(ref_names):
    offsets = {}
    for ref_name in ref_names:
        offsets[ref_name] = 0
    
    return offsets

def create_SNP_id(snp_id_prefix, snps, ref_names):
    digits_no = 7
    snp_no = 1
    for contig_name in ref_names:
        for snp in snps[contig_name]:
            snp.snp_id = snp_id_prefix + str(snp_no).zfill(digits_no)
            snp_no += 1

    digits = int(math.log10(snp_no))+1
    if digits > digits_no:
        print("Error: Increase the amout of digits with the variable digits_no")
        sys.exit(1)


def write_GFF3(file_name, chr_name, snps,ref_names):
    sys.stderr.write( 'Writing ' + file_name + '\n') #+ ' on CPU id ' + str(os.getpid())
    with open(file_name, 'w') as f:
        f.write('##gff-version 3\n')
        for contig_name in ref_names:
            for snp_no, snp in enumerate(snps[contig_name]):
                genotypes = []
                # cul is a string with the cultivar ("T")
                for cul in snp.genotypes:
                    # bases is a Genotype_info with all seen bases and their counts
                    bases = snp.genotypes[cul].bases
                    substring = cul + "="
                    for base in bases:
                        substring += str(bases[base]) + "*" + base + "/"
                    substring = substring[:-1] # remove the last "/"
                    genotypes.append(substring)
                genotypes = sorted(genotypes)
                chr_snp_pos = str(snp.chr_pos + 1)
                p1 = '\t'.join((chr_name, 'SGSautoSNP', 'SNP', chr_snp_pos, chr_snp_pos, '.', '.', '.'))

                snp_name = 'Name=' + snp.snp_id
                snp_id = 'ID=' + snp.snp_id
                this_contig_name = "Contig name=" + contig_name
                
                snp_score = 'SNP score=' + str(snp.snp_score)
                contig_snp_pos = 'SNP pos. on scaffold=' + str(snp.contig_pos + 1)
                gt = 'Genotype ' + ';Genotype '.join(genotypes)
                changes = 'Changes=' + snp.allele
                p2 = ';'.join((snp_name, snp_id, this_contig_name, snp_score, contig_snp_pos, gt, changes))
                line = p1 + '\t' + p2
                f.write(line + '\n')

def write_GFF3m(file_name, snps,ref_names):
    sys.stderr.write( 'Writing ' + file_name + '\n') #+ ' on CPU id ' + str(os.getpid())
    with open(file_name, 'w') as f:
        f.write('##gff-version 3\n')
        for contig_name in ref_names:
            for snp_no, snp in enumerate(snps[contig_name]):
                genotypes = []
                for cul in snp.genotypes:
                    bases = snp.genotypes[cul].bases
                    substring = cul + "="
                    for base in bases:
                        substring += str(bases[base]) + "*" + base + "/"
                    substring = substring[:-1] # remove the last "/"
                    genotypes.append(substring)
                genotypes = sorted(genotypes)
 
                conting_snp_pos = str(snp.contig_pos + 1)
                p1 = '\t'.join((contig_name, 'SGSautoSNP', 'SNP', conting_snp_pos, conting_snp_pos, '.', '.', '.'))

                snp_name = 'Name=' + snp.snp_id
                snp_id = 'ID=' + snp.snp_id
                this_contig_name = "Contig name=" + contig_name

                snp_score = 'SNP score=' + str(snp.snp_score)
                contig_snp_pos = 'SNP pos. on scaffold=' + str(snp.contig_pos + 1)
                gt = 'Genotype ' + ';Genotype '.join(genotypes)
                changes = 'Changes=' + snp.allele
                p2 = ';'.join((snp_name, snp_id, this_contig_name, snp_score, contig_snp_pos, gt, changes))
                line = p1 + '\t' + p2
                f.write(line + '\n')


def write_snp(file_name, snps, ref_names):
    sys.stderr.write( 'Writing ' + file_name + '\n') #+ ' on CPU id ' + str(os.getpid())
    with open(file_name, 'w') as f:
        f.write('\t'.join(('sca.', 'sca. pos', 'chr. pos', 'SNP score', 'genotypes', 'allele', 'SNP id')) + '\n')
        for contig_name in ref_names:
            for snp in snps[contig_name]:
                genotypes = []
                for cul in snp.genotypes:
                    bases = snp.genotypes[cul].bases
                    substring = cul + "="
                    for base in bases:
                        substring += str(bases[base]) + "*" + base + "/"
                    substring = substring[:-1] # remove the last "/"
                    genotypes.append(substring)
                genotypes = sorted(genotypes)
 
                contig_snp_pos = str(snp.contig_pos + 1)
                chr_snp_pos = str(snp.chr_pos + 1)
                snp_score = str(snp.snp_score)
                gt = ';'.join(genotypes)
                changes = snp.allele
                line = '\t'.join((contig_name, contig_snp_pos, chr_snp_pos, snp_score, gt, changes, snp.snp_id))
                f.write(line + '\n')


def write_vcf(file_name, snps, ref_names, fasta_filename, pl, cultivars):
    sys.stderr.write('Writing ' + file_name  + '\n')
    now = datetime.datetime.now()
    with open(file_name, 'w') as f:
        f.write('##fileformat=VCFv4.0\n')
        f.write('##filedate='+now.strftime("%Y%m%d") + '\n')
        f.write('##source=SGSautoSNP\n')
        f.write('##reference='+os.path.basename(fasta_filename) + '\n')
        f.write('##phasing=allhomozygote\n')
        f.write('##INFO=<ID=DP,Number=1,Type=Integer,' +
                'Description="Read depth over all samples">\n')
        f.write('##INFO=<ID=PL,Number=0,Type=String,Description="Panel">\n')

        cultivars_sorted = sorted(cultivars.split(','))
        cultivar_list = '\t'.join(cultivars_sorted)
        f.write('\t'.join(('#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL',
                           'FILTER', 'INFO', 'FORMAT', cultivar_list)) + '\n')
        for contig_name in ref_names:
            for snp in snps[contig_name]:
                coverage = 0
                cul_genotypes = []
                for cul in snp.genotypes:
                    bases = snp.genotypes[cul].bases
                    # bases is now {"A":5}
                    assert len(bases) == 1
                    for base in bases:
                        coverage += bases[base]
                    cul_genotypes.append((cul + "=" + base + "|" + base + ":"+str(bases[base])))
                cul_genotypes = sorted(cul_genotypes)

                snp_allele_bases = snp.allele.split("/")
                snp_allele_split_bases = snp.allele.split("/")

                #contig_snp_pos = str(snp.contig_pos + 1)
                chr_snp_pos = str(snp.chr_pos + 1)
                snp_score = str(snp.snp_score)
                info = 'DP=' + str(coverage) + ';PL=' + pl
                ref_base = snp.ref_base  # contig_seq[snp.chr_pos]
                snp_allele_bases = []
                snp_allele_bases.append(ref_base)
                snp_alt = ''
                for i in snp_allele_split_bases:
                    if i != ref_base:
                        snp_allele_bases.append(i)
                        if snp_alt == '':
                            snp_alt = i
                        else:
                            snp_alt = snp_alt + ',' + i
                genotypes = []
                for cul_genotype in cul_genotypes:
                    genotype_splits = cul_genotype.split("=")
                    genotypes.append(genotype_splits[1])
                gt = '\t'.join(genotypes)
                k = 0
                for j in snp_allele_bases:
                    gt = gt.replace(j, str(k))
                    k = k + 1
                gt = gt.replace("X", ".")
                line = '\t'.join((contig_name, chr_snp_pos, snp.snp_id,
                                  ref_base, snp_alt, snp_score, '.', info,
                                  'GT:DP', gt))
                f.write(line + '\n')


def write_stat(file_name, snps, contig_names):
    sys.stderr.write( 'Writing ' + file_name + '\n') #+ ' on CPU id ' + str(os.getpid())
    snps_no = {}
    mutations = {}

    for contig_name, snps_info in snps.items():
        snps_no[contig_name] = len(snps_info)
        
        #create function and share with Marker
        mutations[contig_name] = {}
        for snp in snps_info:
            try:
                mutations[contig_name][snp.allele] += 1
            except KeyError:
                mutations[contig_name][snp.allele] = 1

    with open(file_name, 'w') as f:
        f.write('\t'.join(('scaffolds', 'SNPs no.', 'mutations')) + '\n')
        for contig_name in contig_names:
            mut = ';'.join(sorted((k + '=' + str(v)) for k,v in mutations[contig_name].items()))
            f.write('\t'.join((contig_name, str(snps_no[contig_name]), mut)) + '\n')

        #create function and share with Marker
        mutations_total = {}
        for vs in mutations.values():
            for k, v in vs.items():
                try:
                    mutations_total[k] += v
                except KeyError:
                    mutations_total[k] = v

        mut_total = ';'.join(sorted((k + '=' + str(v)) for k,v in mutations_total.items()))

        snps_no_total = str(sum([i for i in snps_no.values()]))

        f.write('\t'.join(('Total', snps_no_total, mut_total)) + '\n')


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
            description='SGSautoSNP: a parallel SNP discovery tool for BAM files.')

    parser.add_argument('--bam', nargs='?', required=True,
                       help='bam file with your alignment. Needs a .bai file from "samtools index".')
    
    parser.add_argument('--fasta', nargs='?', required=True,
                   help='Input single/multiple reference fasta file.')

    parser.add_argument('--snp_id_prefix', nargs='?', required=True,
                        help='Please provide the prefix of each SNP ID. eg. \
                        UQ01F')

    parser.add_argument('--pl', nargs='?', required=True,
                        help='Please provide PL for VCF file. eg. "UQ_BNSNP"')

    parser.add_argument('--contig_output', nargs='?', required=True,
                       help='Please provide an output file name. SGSautoSNP ' +\
                            'will attached the following suffix to it: ' +\
                             'gff3, vcf, snp and stat')
    
    parser.add_argument('--chr_offset', nargs='?', required=False,
                       help='Provide an offset file and it will offset contig ' +\
                             'positions on chromosome. You have to use --chr_output, too.')

    parser.add_argument('--chr_output', nargs='?', required=False,
                       help='Please provide an GFF3 output file name for chromosome. ' +\
                             'You have to use --chr_offset, too.')
        
    parser.add_argument('--cultivars', nargs='?', required=True,
                       help='Give all cultivars which are in BAM files eg.' +\
                           ' "J,E,A,S,M1,M2,Bn,Sr"')

    parser.add_argument('--cpu', nargs='?', type=int, required=True, 
                       help='How many CPUs/Cores you would like to use') 

    parser.add_argument('--heterozygous', action='store_true', required=False,
                       help='(Optional) If this flag is set, SNPs which are heterozygous in a cultivar are reported as well.')

    parser.add_argument('-v', '--version', action='version', version='SGSautoSNP %s' %VERSION)
    
    args = parser.parse_args()

    heterozygous = args.heterozygous

    # get all reference names from the fasta file
    #ref_names = [i.id for i in SeqIO.parse(args.fasta, "fasta")] # slow version
    ref_names = get_id_names(args.fasta)
    #ref_info = list((i.id, len(i)) for i in SeqIO.parse(args.fasta, "fasta"))    
    #ref_names = [i[0] for i in ref_info]
    num_jobs = len(ref_names)

    sys.stderr.write('Number of jobs running: ' + str(num_jobs) + '\n')
    
    # Check for existence of .bai file
    if not os.path.exists(args.bam):
        sys.stderr.write("The specified bam-file '" + args.bam + "' does not exist.\n")
        sys.exit(1)

    bai_file = args.bam + ".bai"
    if not os.path.exists(bai_file):
        sys.stderr.write("The bam-file's .bai ('" + args.bam + ".bai') does not exist. Please create one using 'samtools index'\n")
        sys.exit(1)

    # Fix the job count if we have more CPUs than references
    if args.cpu > num_jobs:
        num_cores = num_jobs
        sys.stderr.write('SGSautoSNP uses ' + str(num_cores) + ' cores, because there only this amount of references.\n')
    else:
        num_cores = args.cpu #multiprocessing.cpu_count()

    if args.chr_offset and not args.chr_output:
        sys.stderr.write('The --chr_output option is required whenever --chr_offset is specified\n')
        sys.exit(1)
    elif not args.chr_offset and args.chr_output:
        sys.stderr.write('The --chr_offset option is required whenever --chr_output is specified\n')
        sys.exit(1)  
    
    if args.chr_offset and args.chr_output:
        offsets, chr_name = read_offset_file(args.chr_offset)                
    else:
        offsets = create_fake_offset(ref_names) 
    

    cultivars = args.cultivars.replace(' ', '').replace('"','').replace("'","").split(',')
    if len(set(cultivars)) != len(cultivars):
        sys.stderr.write("FATAL ERROR: the cultivars in your list of cultivars are not unique.\n")
        sys.stderr.write("You gave: " + str(cultivars) + "\n")
        sys.stderr.write("Unique should be: " + str(set(cultivars)) + "\n")
        sys.exit(1)
         
    tasks = JoinableQueue()
    results = Queue()

    # Start Threads       
    sys.stderr.write( 'Creating %d consumers\n' % num_cores)
    threads = [ SNPcaller(tasks, results, args.bam, cultivars, offsets, args.heterozygous, args.fasta)
                  for i in range(num_cores) ]
    
    for i, w in enumerate(threads):
        w.start()
        sys.stderr.write('Thread ' + str(i+1) + ' has started!\n')

    # Enqueue jobs
    for i in ref_names:        
        tasks.put(i)

    # Wait for all of the tasks to finish
    tasks.join()

    snps = {}
    # Start printing results
    for i in range(num_jobs):
        key, value = results.get()
        # did we have an error in a process?
        if value == False:
            sys.exit(1)
        assert key not in snps
        snps[key] = value

    # Add None to kill each thread
    for i in range(num_cores):
        tasks.put(None)

    # Overwrite all SNP ids - this is slow!    
    create_SNP_id(args.snp_id_prefix, snps, ref_names)
        
    if args.chr_offset and args.chr_output:
        write_GFF3(args.chr_output, chr_name, snps, ref_names)
        vcf_filename = args.chr_output + '.vcf'
    else:
        vcf_filename = args.contig_output + '.vcf'
    
    # write all output
    gff3m_filename = args.contig_output + '_snp.gff3'
    snp_filename = args.contig_output + '.snp'
    stat_filename = args.contig_output + '_snp.stat' 
    
    write_GFF3m(gff3m_filename, snps, ref_names)
    if heterozygous:
        sys.stderr.write("Checking for heterozygous SNPs, won't write .VCF.\n")
    else:
        write_vcf(vcf_filename, snps, ref_names, args.fasta, args.pl, args.cultivars)
    write_snp(snp_filename, snps, ref_names)
    write_stat(stat_filename, snps, ref_names)
