#!/usr/bin/perl

use strict;
use warnings;
use IO::Handle;
use Getopt::Long;
use Pod::Usage;
use Data::Dumper;
use File::Basename;

BEGIN {
	STDERR->autoflush(1);
	STDOUT->autoflush(1);
}


########## options ##########
my ( $help, $man, $DEBUG, $outFilename, $noHeader );
my $delimit=",";
my $markerSym="M";

GetOptions(
		'help|?' => \$help, 
		'man' => \$man,
		'debug' =>\$DEBUG,
		'out=s' => \$outFilename,
		'delimit|delimiter=s' => \$delimit,
		'noHeader' => \$noHeader,
		'markerSymbol=s' => \$markerSym,
	) or pod2usage(2);
pod2usage(1) if $help;
pod2usage(-verbose => 2) if $man;


########## arguments ##########
pod2usage("***** Error: Incorrect num of arguments *****") if (@ARGV <= 0);
pod2usage("***** Error: please specify output filename *****") if (!$outFilename);


########## sub-routine to add commas to numbers ##########
sub commify {
    # example: 123456 -> 123,456
    my $number = shift;
    $number = reverse $number;
    $number =~ s<(\d\d\d)(?=\d)(?!\d*\.)><$1,>g;
    return reverse $number;
}

########## program start ##########
&printTime("Start");

my ($outMap, $outGenotype);
# open 2 output file handle: MAP & GENOTYPE
open($outMap, ">$outFilename.map") || die $!;
open($outGenotype, ">$outFilename.genotype") || die $!;

print $outMap "# fjFile = MAP\n";
print $outGenotype "# fjFile = GENOTYPE\n";

my ($curInFile);
my $curLine;
my $markerNumOfDigit;
my (@fields, $markerID);

# go through each input files
for (my $i=0; $i<@ARGV; $i++) {

	# if it's first file go through the file to print the marker id line
	if ($i == 0) {
		# open input file handle
		open($curInFile, "<$ARGV[$i]") || die $!;
		while (<$curInFile>) {
			$curLine=$_ if eof;
		}
		@fields = split($delimit, $curLine);
		$markerID = $fields[0];
		$markerNumOfDigit=length($markerID);

		# open input file handle
		open($curInFile, "<$ARGV[$i]") || die $!;
		# take out the header line
		$curLine=<$curInFile> if (!$noHeader);

	  # write MarkerID line to GENOTYPE
		while ($curLine=<$curInFile>) {
			# collect fields based on delimiter
			@fields = split($delimit, $curLine);
			$markerID = $fields[0];
			my ($chrID, $chrPos) = ($fields[1], $fields[4]);
            $chrPos = commify($chrPos);

			printf $outGenotype "\t$markerSym$markerID";

			# write MarkerID\tChrID\tChrPOS to MAP
			print $outMap "$markerSym$markerID\t$chrID\t$chrPos\n";
		}
		print $outGenotype "\n";

	}

	my @temp=split('_', basename($ARGV[$i]));
	my $cultivarID=$temp[1];
	print $outGenotype "$cultivarID";

	# open input file handle
	open($curInFile, "<$ARGV[$i]") || die $!;
	# take out the header line
	$curLine=<$curInFile> if (!$noHeader);

	# go through each line
	while ($curLine=<$curInFile>) {
		chomp $curLine;
		# collect fields based on delimiter
		@fields = split($delimit, $curLine);
		# field0=markerID, field1=ChrID, field4=ChrPOS, field5=Genotype
		my $genotype=$fields[5];
        if ($genotype eq "-") {
            $genotype = "";
        }

	  # write \tSampleID to GENOTYPE
		print $outGenotype "\t$genotype";
	}

	print $outGenotype "\n";
}


&printTime("End");


########## functions ##########
sub printTime {
	my $message;
	if (@_ >= 1) {
		($message) = @_;
	} else {
		$message = "Current Time";
	}
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime(time);
	printf "$message: %4d-%02d-%02d %02d:%02d:%02d\n",$year+1900,$mon+1,$mday,$hour,$min,$sec; 
}


__DATA__

=head1 NAME

parseGenotype.pl - This script will parse the genotype files and output them to a single flapjack import file set (.genotype and .map files).


=head1 SYNOPSIS

parseGenotype.pl [-help] [-man] [-debug] -out <output_filename> [-delimiter <CHAR>] [-noHeader] [-markerSymbol <STRING>] <genotype_csv_file1> [<genotype_csv_file2> ...]

B<Example>

parseGenotype.pl -out GBS_IsobellC8_82cultivar GBS_IsobellC8/*_csv


=head1 OPTIONS

=over 8

=item B<-help>

Print a brief help message and exits.

=item B<-man>

Prints the manual page and exits.

=item B<-debug>

Turn on the script debugging printout.

=item B<-out>

Output file name

=item B<-delimiter>

The delimiter used in the input file. Default is ','

=item B<-noHeader>

Set it when the genotype csv files does not have a header

=item B<-markerSymbol>

The string or char symbol used to describe the marker name

=back


=head1 VERSION

	Version 1.0
		Initial version

=head1 LICENSE

	Created by: Kenneth Chan (Version 1.0)
    Modified by: Philipp Bayer (Version 1.1, changed first line of each file to "official" Flapjack line, replaced "-"-alleles by "", added commify-function)

=cut

