#!/usr/bin/env perl

use strict;
use warnings;

use Bio::DB::Sam;
use Bio::SeqIO;
use Carp;
use Getopt::Long;
use File::Basename;

BEGIN {
	*STDOUT->autoflush();
	*STDERR->autoflush();
}

my $options = check_params();

my @snp_order          = ();
my %snp_details        = ();
my %genotyping_details = ();
my %offsets            = ();
my @tabular_details    = ();
my $total_hybrid_snps  = 0;
my $count              = 0;
my $snp_contig_pos;
my $snp_id;
my $genotype;
my $hybrid_bases;
my ($chr_id) = retrieve_chr_id();

my $basename = basename( $options->{'gff'}, '.gff3' );

$basename = $options->{'o'} if $options->{'o'};


my $basename_cultivars = $basename . '_' . $options->{'c1'} . $options->{'c2'} ;
my $genotyping_output_file = $basename_cultivars . '.genotyping';
open my $genotyping_fh, '>', $genotyping_output_file
  or croak("Can't open output file [$genotyping_output_file]: $!.\n");

retrieve_contigs_offsets( \%offsets );
retrieve_snp_details_from_gff3();
process_hybrid_snps();
print_summary();

close $genotyping_fh
  or croak("Failed to close [$genotyping_output_file]: $!\n");

#------------------------------------------------------------------
sub print_summary {
	my $total_cultivars_gff3_snps = @snp_order;
	print "***\n";
	print "Total snps between cultivars [$options->{'c1'}] and [$options->{'c2'}] in gff3 file [$options->{'gff'}] [$total_cultivars_gff3_snps]\n";
	print "Total snps between cultivars [$options->{'c1'}] and [$options->{'c2'}] with aligned reads in hybrid bam file [$options->{'b'}] [$total_hybrid_snps]\n";
	print "***\n";
}

sub retrieve_snp_details_from_gff3 {
	open my $gff_fh, '<', $options->{'gff'}
	  or croak("Can't open input file [$options->{'gff'}]: $!.\n");

	while ( my $line = <$gff_fh> ) {
		next if $line =~ m/\A \# /xms;

		my @feature_values = split( "\t", $line );
		my $snp_desc = $feature_values[8];

		my ( $c1_base_count, $c1_base_type ) =
		  retrieve_base_type_and_count( $snp_desc, $options->{'c1'} );
		my ( $c2_base_count, $c2_base_type ) =
		  retrieve_base_type_and_count( $snp_desc, $options->{'c2'} );
		my $is_snp_valid = validate_snp($c1_base_count, $c1_base_type, $c2_base_count, $c2_base_type);

		next if !$is_snp_valid;

		my ($snp_name) = $snp_desc =~ m/\A Name= ([^;]+);/xms;
		push @snp_order, $snp_name;

		my $contig_id = $feature_values[0];
		my $snp_pos   = $feature_values[3];

		$snp_details{$snp_name} =
		  [ $contig_id, $snp_pos, $line, $c1_base_type, $c2_base_type, ];
	}

	close $gff_fh or croak("Failed to close [$options->{'gff'}]: $!\n");
}

=doc
Checks for a given SNP, and a pair of genotypes on that SNP, whether the SNP in question has to be discarded.
Rules: 
1. Both genotypes' alleles can't be missing (!= "X")
2. Both alleles have to be different
3. Both counts have to be higher than the minimum score (given by option 's')
Returns 0 if improper SNP or 1 if proper SNP.
=cut
sub validate_snp {
	my ( $c1_base_count, $c1_base_type, $c2_base_count, $c2_base_type ) = @_;

	if ( (( $c1_base_type eq 'X' ) || ( $c2_base_type eq 'X' ))
		  || 
          ( $c1_base_type eq $c2_base_type )
          || 
          (( $c1_base_count < $options->{'s'} ) || ( $c2_base_count < $options->{'s'} ))
	   ) {
		return 0;
	}
	else { 
        return 1;
    }
}

=doc
Receives a part of the SNP gff3 file (the last column) and the cultivar in question.
It returns the number of alleles ($base_count) and the genotype of that cultivar ($base_type).
Example:
$snp_desc="Name=Ca00000000009;ID=Ca00000000009;Contig name=Ca9;SNP score=11;SNP pos. on scaffold=1724;Genotype AA=7*G;Genotype AB=4*G;Genotype AC=3*G;Genotype AD=8*G;Genotype AE=3*G;Genotype AF=2*G;"
$cultivar="AF"
returns $base_count=2 and $base_type=G
=cut
sub retrieve_base_type_and_count {
	my ( $snp_desc, $cultivar ) = @_;
    my $base_count = 0;
    my $base_type = "";
    my @types_array = split(';', $snp_desc);
    foreach my $genotype (@types_array) {
        if (($genotype =~ m/Genotype/) && ($genotype =~ m/$cultivar/)) {
            my @sub_list = split("=",$genotype); # make "Genotype AP=1*A/2*G" into ["Genotype AP", "1*A/2*G"], get only last
            my $sub_list = $sub_list[1];
            my @all_nums = $sub_list =~ /(\d+)/g; # 1*A/2*G becomes (1,2), 1*A becomes (1)
            my @all_letters = $sub_list =~ /([A-z]+)/g;  # 1*A/2*G becomes (A,G), 2*G becomes (G)
            # sort
            @all_letters = sort @all_letters;
            $base_type = join('', @all_letters);
            foreach my $count (@all_nums) {
                $base_count = $base_count + $count;
            }
        }
    }

    ## OLD: works only with homozygous SNPs, above new code works with hetero/homozygous
    #my ( $base_count, $base_type ) =
    #$snp_desc =~ m/Genotype [ ]+ $cultivar=(\d+) \* (\w); /xms;

	return $base_count, $base_type;
}

sub process_hybrid_snps {
	my $sam = Bio::DB::Sam->new(
		-bam   => $options->{'b'},
		-fasta => $options->{'f'},
	);

	my $tabular_output_file =$basename_cultivars . '.genotyping_csv';
	open my $tabular_fh, '>', $tabular_output_file
	  or croak("Can't open output file [$tabular_output_file]: $!.\n");

	my $header = "No,Chr Id,SNP Id,Contig Pos,Chr Pos,Genotype,Pileup";
	print $tabular_fh $header;

	foreach my $id (@snp_order) {
		$snp_id = $id;
		my $snp_id_details_ref = $snp_details{$snp_id};
		my $contig             = $snp_id_details_ref->[0];
		$count++;

		$snp_contig_pos = $snp_id_details_ref->[1];
		my $offset      = $offsets{$contig};
		my $snp_chr_pos = $snp_contig_pos + $offset;
		my $region = $contig . ':' . $snp_contig_pos . '-' . $snp_contig_pos;

		@tabular_details = ();
		push @tabular_details, $count, $chr_id, $snp_id, $snp_contig_pos,
		  $snp_chr_pos;

		## N as default is confusing, changing it to -
		# $genotype     = 'N';
        $genotype     = '-';
		$hybrid_bases = '';
		$sam->fast_pileup( $region, \&snp_caller );
		push @tabular_details, $genotype;
		push @tabular_details, $hybrid_bases if $hybrid_bases;

		print_csv_line( $tabular_fh, \@tabular_details );
	}

	close $tabular_fh or croak("Failed to close [$tabular_output_file]: $!\n");
}

sub print_csv_line {
	my ( $tabular_fh, $tabular_details_ref ) = @_;

	my $line_values = join( ',', @$tabular_details_ref );
	print $tabular_fh "\n$line_values";
}

sub snp_caller {
	my ( $seqid, $pos, $p ) = @_;

	return if ( $pos != $snp_contig_pos );

	my @snp_bases = ();
	my @reads_ids = ();
	my $cultivar  = $options->{'c1'};
    #my ($base_type) =
    #  $snp_details{$snp_id}[2] =~ m/Genotype [ ]+ $cultivar=\d+ \* (\w); /xms;
    #
    my $base_type;
    my @types_array = split(';', $snp_details{$snp_id}[2]);
    foreach my $genotype (@types_array) {
        if (($genotype =~ m/Genotype/) && ($genotype =~ m/$cultivar/)) {
            my @sub_list = split("=",$genotype); # make "Genotype AP=1*A/2*G" into ["Genotype AP", "1*A/2*G"], get only last
            my $sub_list = $sub_list[1];
            my @all_nums = $sub_list =~ /(\d+)/g; # 1*A/2*G becomes (1,2), 1*A becomes (1)
            my @all_letters = $sub_list =~ /([A-z]+)/g;  # 1*A/2*G becomes (A,G), 2*G becomes (G)
            $base_type = join('', @all_letters);
        }
    }

	for my $pileup (@$p) {
		next if ( $pileup->indel );

		my $b       = $pileup->alignment;
		my $read_id = $b->query->name;
		my $qbase   = substr( $b->qseq, $pileup->qpos, 1 );
		next if $qbase =~ /[nN]/;
		push @snp_bases, $qbase;
		push @reads_ids, $read_id if ( $base_type ne $qbase );
	}

	push @{ $snp_details{$snp_id} }, \@snp_bases;
	my ($c1_base_type) = $snp_details{$snp_id}[3];
	my ($c2_base_type) = $snp_details{$snp_id}[4];

	$genotype = determine_genotype( $c1_base_type, $c2_base_type, \@snp_bases );
	$hybrid_bases = join( '', @snp_bases );

	my $sorted_bases_ref = sort_snp_bases( \@snp_bases );
	my $bases_count_line = count_bases($sorted_bases_ref);
	print_genotyping_line( \@snp_bases, $bases_count_line );
	$total_hybrid_snps++;
}

sub print_genotyping_line {
	my ( $snp_bases_ref, $bases_count_line ) = @_;

	print $genotyping_fh $snp_details{$snp_id}[2];
	print $genotyping_fh "\t *** Hybrid cultivar $options->{'h'} *** ",
	  @$snp_bases_ref, "\t$bases_count_line\n";
}

sub retrieve_chr_id {
	open my $contigs_offset_fh, '<', $options->{'off'}
	  or croak("Can't open contigs offset file [$options->{'off'}]: $!.\n");
	my $chr_id;

	while ( my $line = <$contigs_offset_fh> ) {
		next if ( $line =~ m/\A \# /xms );
		my @feature_values = split( "\t", $line );
		$chr_id = $feature_values[0];

		last;
	}

	close $contigs_offset_fh
	  or croak("Failed to close [$options->{'off'}]: $!\n");

	return $chr_id;
}

sub retrieve_contigs_offsets {
	my ($offsets_ref) = @_;

	open my $contigs_offset_fh, '<', $options->{'off'}
	  or croak("Can't open contigs offset file [$options->{'off'}]: $!.\n");

	while ( my $line = <$contigs_offset_fh> ) {
		next if ( $line =~ m/\A \# /xms );
		my @gff_columns = split( "\t", $line );
		my $offset      = ( $gff_columns[3] - 1 );
		my $notes       = $gff_columns[8];
		my ($id) = $notes =~ m/ID= ([^;]+) ; /xms;
		$offsets_ref->{$id} = $offset;
	}

	close $contigs_offset_fh
	  or croak("Failed to close [$options->{'off'}]: $!\n");
}

sub is_conflict {
	my ( $sorted_bases_ref, $base_type ) = @_;

	my $conflict = 0;

	foreach my $base ( sort keys %$sorted_bases_ref ) {
		$conflict++ if !( $base eq $base_type );
	}

	return $conflict;
}

sub determine_genotype {
	my ( $c1_base_type, $c2_base_type, $chars_ref ) = @_;

	my %bases = ();
	foreach my $char (@$chars_ref) {
		$bases{$char}++;
	}

	my $bases_count = keys %bases;

    # return "HETEROZYGOUS" if more than 2 alleles found at this position
	return 'E' if ( $bases_count > 1 );

	my ($base) = keys %bases;

    # new rules: child can still have "A" as base while both parents can be "AG" and "A", 
    # should be heterozygous then
    if ((index($c1_base_type, $base) != -1) && (index($c2_base_type, $base) != -1)) {
        return 'E';
    }

    return $options->{'c1'} if ( $c1_base_type eq $base );
    return $options->{'c2'} if ( $c2_base_type eq $base );

	return 'E';
}

sub count_bases {
	my ($sorted_bases_ref) = @_;

	my @bases_count = ();

	foreach my $base ( sort keys %$sorted_bases_ref ) {
		my $base_total = $sorted_bases_ref->{$base};
		push @bases_count, $base_total . '*' . $base;
	}

	my $bases_count_string = join( ', ', @bases_count );

	return $bases_count_string;
}

sub sort_snp_bases() {
	my ($snp_bases_ref) = @_;

	my %sorted_bases = ();

	foreach my $base (@$snp_bases_ref) {
		$sorted_bases{$base}++;
	}

	return \%sorted_bases;
}

sub check_params {
	my @standard_options = ( "help+", "man+" );
	my %options;

	# Add any other command line options, and the code to handle them
	GetOptions(
		\%options, @standard_options, "gff:s", "b:s",
		"f:s",     "c1:s",            "c2:s",  "off:s",
		"h:s",     "s:s",
		"o:s"
	);

	exec("pod2usage $0") if $options{'help'};
	exec("perldoc $0")   if $options{'man'};
	exec("pod2usage $0")
	  if !(
			     $options{'gff'}
			  && $options{'b'}
			  && $options{'f'}
			  && $options{'c1'}
			  && $options{'c2'}
			  && $options{'h'}
			  && $options{'off'}
			  && $options{'s'}
	  );

	return \%options;
}

__DATA__

=back 

=head1 NAME

   snp_genotyping_all.pl

=head1 VERSION

    Version 1.1

        Script can now handle heterozygous input alleles. 

	Version 1.01

		Changed default genotype symblo from 'N' to '-'
		Added '-o' option to accept output file name


	Version 1.0

		Originial version

=head1 COPYRIGHT

    Further by: Philipp Bayer (v1.1) 
	Modify by: Kenneth Chan (v1.01)
	Create by: Jiri Stiller (v1.0)

  Copyright (c) 2011 Jiri Stiller, 2012 Kenneth Chan, 2014 Philipp Bayer

=head1 DESCRIPTION
   
   The script retrieves bases for each SNP from hybrid line bam file 
   and generates a new .genotyping snp file with added hybrid information. 
   It also creates a .genotyping_csv file containing tabular output including genotype 
   and offset position for each snp.


=head1 SYNOPSIS

   snp_genotyping_all.pl -gff <gff3_file_name> -b <bam_file_name> -f <fasta_file_name> -off <gf3_file_name> -c1 <cultivar_1> -c2 <cultivar_2> -h <hybrid_name> -s 1 [-help] [-man] [-o <out_file_name>]

      [-help]     Displays basic usage information
      [-man]      Displays more detailed information
      -gff        Input gff3 file containg snp positions on contigs 
      -b          Bam file
      -f          Fasta file
      -off        Chromosome contig offset file
      -c1         Cultivar name
      -c2         Cultivar name
      -h		  Hybrid name
      -s          minimum score for each cultivar
      -o          output file name, if not specify, the snp file name will be used
      
   Examples:
   snp_genotyping_all.pl -b A2_37_001_vs_BAm_V4.0_BCm_V2.0.paired.sorted.bam -gff J_E_A_S_M1_M2_Bn_Sr_vs_BA01m_v4.0_snp.gff3 -off BA01_v4.0_contig.gff3 -f BA01m_v4.0.fasta -c1 A -c2 S -s 1 -h HYBRID
   
   Output:
   J_E_A_S_M1_M2_Bn_Sr_vs_BA01m_v4.0_snp_AS.gff3.genotyping
   J_E_A_S_M1_M2_Bn_Sr_vs_BA01m_v4.0_snp_AS.gff3_genotyping_csv
                                   
=cut
