#!/usr/bin/env python
""" Imputes Flapjack dat files by the rules defined in impute().

Author: Philipp Bayer, MIT License
"""

VERSION = 1.1

def impute(linelist):
    ''' 
    imputes alleles according to the following rules:
    1. if empty alleles are bounded by identical alleles, fill up with these:
        T, "", "", "", T -> T, t, t, t, T
    2. If empty alleles are bounded by different alleles, leave
        T, "", "", "", N -> T, "", "", "", N
    3. If start of end of allele-list, leave
        "", "", "", T -> "", "", "", T
    '''

    from collections import deque
    new_list = deque([linelist[0]])
    index = 1 # element 0 is TNXX
    start = ""

    inside = False
    counter = 0
    while index < len(linelist):
        current = linelist[index]

        if current == "":
            # inside an empty block
            if inside == False and index != 1:
                start = linelist[index-1]
                counter = 0
            counter += 1
            inside = True
        elif current != "" and inside == True:
            # finished an empty block
            end = current
            if start == end:
                
                #to_replace_with = string.ascii_letters[string.ascii_letters.index(end) + 1]
                to_replace_with = end.lower() 
                new_list += counter * [to_replace_with] 
            elif start != end:
                new_list += counter * [""]
            inside = False
            new_list.append(current)
        elif current != "" and inside == False:
            # if we're inside a block of alleles, just continue
            new_list.append(current)

        if index == len(linelist)-1 and current == "":
            # no end when the list ends in ""!
            new_list += counter * [""]
        index += 1

    assert(len(new_list) == len(linelist))
    return new_list

def iterate(fh, output_fh):
    ''' iterates over lines in the genotype-file '''
    for line in fh:
        linelist = line.replace("\n","").upper().split("\t")
        new_list = impute(linelist)
        output_fh.write("\t".join(new_list) + "\n")

if __name__ == "__main__":

    import argparse
    import sys

    parser = argparse.ArgumentParser(description="Parses an exported Flapjack-dat file. Imputes missing alleles, if the missing alleles are bordered by identical alleles.")

    parser.add_argument("dat", help="Name of the dat-file")

    args = parser.parse_args()

    assert ".dat" in args.dat, "ERROR: Given file is not a .dat file"
    fh = open(args.dat)

    extra = fh.readline()
    # check whether this file is correct format
    if "# fjFile = GENOTYPE" not in extra:
        sys.stderr.write("Error: File doesn't seem to be a Flapjack dat-file.\n")
        sys.exit(1)

    output = args.dat.replace(".dat","_imputed.dat").replace(" ","_")
    output_fh = open(output, "w")
    sys.stderr.write("Writing output to %s\n" %(output))

    header = fh.readline() # leave this unchanged
    output_fh.write(extra)
    output_fh.write(header)

    iterate(fh, output_fh)

    sys.stderr.write("Done!\n")
